#pragma once
#ifndef CATA_SRC_MONEXAMINE_H
#define CATA_SRC_MONEXAMINE_H

class monster;

namespace monexamine
{

bool pet_menu( monster &z );
bool mech_hack( monster &z );
bool pay_bot( monster &z );
bool mfriend_menu( monster &z );

} // namespace monexamine

#endif // CATA_SRC_MONEXAMINE_H
